"""
Enhanced error types for resource management.

This module implements comprehensive error types and error context building
for production-ready resource management, providing detailed error information
and context for debugging and monitoring.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
import time
from typing import Any

import multiaddr

from nanolib.peer.id import ID


class ErrorSeverity(Enum):
    """Error severity levels."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


class ErrorCategory(Enum):
    """Error categories for classification."""

    # Resource limit errors
    CONNECTION_LIMIT = "connection_limit"
    MEMORY_LIMIT = "memory_limit"
    STREAM_LIMIT = "stream_limit"
    PEER_LIMIT = "peer_limit"
    PROTOCOL_LIMIT = "protocol_limit"

    # System errors
    SYSTEM_ERROR = "system_error"
    NETWORK_ERROR = "network_error"
    TRANSPORT_ERROR = "transport_error"
    PROTOCOL_ERROR = "protocol_error"

    # Configuration errors
    CONFIG_ERROR = "config_error"
    VALIDATION_ERROR = "validation_error"

    # Operational errors
    OPERATIONAL_ERROR = "operational_error"
    MONITORING_ERROR = "monitoring_error"
    LIFECYCLE_ERROR = "lifecycle_error"


class ErrorCode(Enum):
    """Standardized error codes."""

    # Connection limit errors (1000-1099)
    CONNECTION_LIMIT_EXCEEDED = "CONN_001"
    CONNECTION_PENDING_LIMIT_EXCEEDED = "CONN_002"
    CONNECTION_ESTABLISHED_LIMIT_EXCEEDED = "CONN_003"
    CONNECTION_PER_PEER_LIMIT_EXCEEDED = "CONN_004"
    CONNECTION_TOTAL_LIMIT_EXCEEDED = "CONN_005"

    # Memory limit errors (1100-1199)
    MEMORY_LIMIT_EXCEEDED = "MEM_001"
    MEMORY_PROCESS_LIMIT_EXCEEDED = "MEM_002"
    MEMORY_SYSTEM_LIMIT_EXCEEDED = "MEM_003"
    MEMORY_MONITORING_FAILED = "MEM_004"

    # Stream limit errors (1200-1299)
    STREAM_LIMIT_EXCEEDED = "STR_001"
    STREAM_PER_CONNECTION_LIMIT_EXCEEDED = "STR_002"
    STREAM_PER_PEER_LIMIT_EXCEEDED = "STR_003"
    STREAM_TOTAL_LIMIT_EXCEEDED = "STR_004"

    # Peer limit errors (1300-1399)
    PEER_LIMIT_EXCEEDED = "PEER_001"
    PEER_CONNECTION_LIMIT_EXCEEDED = "PEER_002"
    PEER_STREAM_LIMIT_EXCEEDED = "PEER_003"

    # Protocol limit errors (1400-1499)
    PROTOCOL_LIMIT_EXCEEDED = "PROT_001"
    PROTOCOL_STREAM_LIMIT_EXCEEDED = "PROT_002"
    PROTOCOL_RATE_LIMIT_EXCEEDED = "PROT_003"

    # System errors (2000-2099)
    SYSTEM_ERROR = "SYS_001"
    SYSTEM_MONITORING_FAILED = "SYS_002"
    SYSTEM_RESOURCE_UNAVAILABLE = "SYS_003"

    # Network errors (2100-2199)
    NETWORK_ERROR = "NET_001"
    NETWORK_CONNECTION_FAILED = "NET_002"
    NETWORK_TIMEOUT = "NET_003"
    NETWORK_UNREACHABLE = "NET_004"

    # Transport errors (2200-2299)
    TRANSPORT_ERROR = "TRANS_001"
    TRANSPORT_CONNECTION_FAILED = "TRANS_002"
    TRANSPORT_STREAM_FAILED = "TRANS_003"
    TRANSPORT_PROTOCOL_ERROR = "TRANS_004"

    # Configuration errors (3000-3099)
    CONFIG_ERROR = "CFG_001"
    CONFIG_VALIDATION_FAILED = "CFG_002"
    CONFIG_INVALID_LIMITS = "CFG_003"
    CONFIG_MISSING_REQUIRED = "CFG_004"

    # Operational errors (4000-4099)
    OPERATIONAL_ERROR = "OP_001"
    OPERATIONAL_MONITORING_FAILED = "OP_002"
    OPERATIONAL_LIFECYCLE_FAILED = "OP_003"
    OPERATIONAL_EVENT_FAILED = "OP_004"


@dataclass
class ErrorContext:
    """Context information for errors."""

    # Basic error information
    error_code: ErrorCode
    error_category: ErrorCategory
    severity: ErrorSeverity
    message: str
    timestamp: float = field(default_factory=time.time)

    # Resource information
    resource_type: str | None = None
    resource_id: str | None = None
    resource_scope: str | None = None

    # Connection information
    connection_id: str | None = None
    peer_id: ID | None = None
    local_addr: multiaddr.Multiaddr | None = None
    remote_addr: multiaddr.Multiaddr | None = None

    # Limit information
    limit_type: str | None = None
    limit_value: int | float | None = None
    current_value: int | float | None = None
    limit_percentage: float | None = None

    # System information
    system_memory_total: int | None = None
    system_memory_available: int | None = None
    system_memory_percent: float | None = None
    process_memory_bytes: int | None = None
    process_memory_percent: float | None = None

    # Additional context
    metadata: dict[str, Any] = field(default_factory=dict)
    stack_trace: str | None = None
    previous_errors: list[ErrorContext] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        """Convert error context to dictionary."""
        return {
            "error_code": self.error_code.value,
            "error_category": self.error_category.value,
            "severity": self.severity.value,
            "message": self.message,
            "timestamp": self.timestamp,
            "resource_type": self.resource_type,
            "resource_id": self.resource_id,
            "resource_scope": self.resource_scope,
            "connection_id": self.connection_id,
            "peer_id": str(self.peer_id) if self.peer_id else None,
            "local_addr": str(self.local_addr) if self.local_addr else None,
            "remote_addr": str(self.remote_addr) if self.remote_addr else None,
            "limit_type": self.limit_type,
            "limit_value": self.limit_value,
            "current_value": self.current_value,
            "limit_percentage": self.limit_percentage,
            "system_memory_total": self.system_memory_total,
            "system_memory_available": self.system_memory_available,
            "system_memory_percent": self.system_memory_percent,
            "process_memory_bytes": self.process_memory_bytes,
            "process_memory_percent": self.process_memory_percent,
            "metadata": self.metadata,
            "stack_trace": self.stack_trace,
            "previous_errors": [err.to_dict() for err in self.previous_errors],
        }

    def __str__(self) -> str:
        """String representation of error context."""
        return (
            f"{self.error_code.value}: {self.message} "
            f"(severity={self.severity.value}, category={self.error_category.value})"
        )


class ResourceLimitExceededError(Exception):
    """Enhanced exception for resource limit exceeded errors."""

    def __init__(
        self,
        error_context: ErrorContext,
        original_exception: Exception | None = None,
    ):
        """
        Initialize resource limit exceeded error.

        Args:
            error_context: Detailed error context
            original_exception: Original exception that caused this error

        """
        self.error_context = error_context
        self.original_exception = original_exception

        # Build error message
        message = self._build_error_message()
        super().__init__(message)

    def _build_error_message(self) -> str:
        """Build detailed error message."""
        ctx = self.error_context

        # Base message
        message = f"{ctx.error_code.value}: {ctx.message}"

        # Add limit information
        if ctx.limit_type and ctx.limit_value is not None:
            if ctx.current_value is not None:
                message += f" (limit={ctx.limit_value}, current={ctx.current_value})"
            else:
                message += f" (limit={ctx.limit_value})"

        # Add resource information
        if ctx.resource_type:
            message += f" [resource={ctx.resource_type}"
            if ctx.resource_id:
                message += f":{ctx.resource_id}"
            message += "]"

        # Add connection information
        if ctx.connection_id:
            message += f" [connection={ctx.connection_id}]"

        # Add peer information
        if ctx.peer_id:
            message += f" [peer={ctx.peer_id}]"

        return message

    def get_error_summary(self) -> dict[str, Any]:
        """Get error summary for monitoring."""
        return {
            "error_code": self.error_context.error_code.value,
            "error_category": self.error_context.error_category.value,
            "severity": self.error_context.severity.value,
            "message": self.error_context.message,
            "timestamp": self.error_context.timestamp,
            "resource_type": self.error_context.resource_type,
            "limit_type": self.error_context.limit_type,
            "limit_value": self.error_context.limit_value,
            "current_value": self.error_context.current_value,
            "has_original_exception": self.original_exception is not None,
        }

    def __str__(self) -> str:
        """String representation of the error."""
        return self._build_error_message()


class SystemResourceError(Exception):
    """Enhanced exception for system resource errors."""

    def __init__(
        self,
        error_context: ErrorContext,
        original_exception: Exception | None = None,
    ):
        """
        Initialize system resource error.

        Args:
            error_context: Detailed error context
            original_exception: Original exception that caused this error

        """
        self.error_context = error_context
        self.original_exception = original_exception

        # Build error message
        message = self._build_error_message()
        super().__init__(message)

    def _build_error_message(self) -> str:
        """Build detailed error message."""
        ctx = self.error_context

        # Base message
        message = f"{ctx.error_code.value}: {ctx.message}"

        # Add system information
        if ctx.system_memory_total is not None:
            message += f" [system_memory={ctx.system_memory_total} bytes"
            if ctx.system_memory_percent is not None:
                message += f" ({ctx.system_memory_percent:.1f}%)"
            message += "]"

        # Add process information
        if ctx.process_memory_bytes is not None:
            message += f" [process_memory={ctx.process_memory_bytes} bytes"
            if ctx.process_memory_percent is not None:
                message += f" ({ctx.process_memory_percent:.1f}%)"
            message += "]"

        return message

    def get_error_summary(self) -> dict[str, Any]:
        """Get error summary for monitoring."""
        return {
            "error_code": self.error_context.error_code.value,
            "error_category": self.error_context.error_category.value,
            "severity": self.error_context.severity.value,
            "message": self.error_context.message,
            "timestamp": self.error_context.timestamp,
            "system_memory_total": self.error_context.system_memory_total,
            "system_memory_percent": self.error_context.system_memory_percent,
            "process_memory_bytes": self.error_context.process_memory_bytes,
            "process_memory_percent": self.error_context.process_memory_percent,
            "has_original_exception": self.original_exception is not None,
        }

    def __str__(self) -> str:
        """String representation of the error."""
        return self._build_error_message()


class ConfigurationError(Exception):
    """Enhanced exception for configuration errors."""

    def __init__(
        self,
        error_context: ErrorContext,
        original_exception: Exception | None = None,
    ):
        """
        Initialize configuration error.

        Args:
            error_context: Detailed error context
            original_exception: Original exception that caused this error

        """
        self.error_context = error_context
        self.original_exception = original_exception

        # Build error message
        message = self._build_error_message()
        super().__init__(message)

    def _build_error_message(self) -> str:
        """Build detailed error message."""
        ctx = self.error_context

        # Base message
        message = f"{ctx.error_code.value}: {ctx.message}"

        # Add configuration information
        if ctx.metadata.get("config_key"):
            message += f" [config_key={ctx.metadata['config_key']}]"

        if ctx.metadata.get("config_value"):
            message += f" [config_value={ctx.metadata['config_value']}]"

        return message

    def get_error_summary(self) -> dict[str, Any]:
        """Get error summary for monitoring."""
        return {
            "error_code": self.error_context.error_code.value,
            "error_category": self.error_context.error_category.value,
            "severity": self.error_context.severity.value,
            "message": self.error_context.message,
            "timestamp": self.error_context.timestamp,
            "config_key": self.error_context.metadata.get("config_key"),
            "config_value": self.error_context.metadata.get("config_value"),
            "has_original_exception": self.original_exception is not None,
        }

    def __str__(self) -> str:
        """String representation of the error."""
        return self._build_error_message()


class OperationalError(Exception):
    """Enhanced exception for operational errors."""

    def __init__(
        self,
        error_context: ErrorContext,
        original_exception: Exception | None = None,
    ):
        """
        Initialize operational error.

        Args:
            error_context: Detailed error context
            original_exception: Original exception that caused this error

        """
        self.error_context = error_context
        self.original_exception = original_exception

        # Build error message
        message = self._build_error_message()
        super().__init__(message)

    def _build_error_message(self) -> str:
        """Build detailed error message."""
        ctx = self.error_context

        # Base message
        message = f"{ctx.error_code.value}: {ctx.message}"

        # Add operational information
        if ctx.metadata.get("operation"):
            message += f" [operation={ctx.metadata['operation']}]"

        if ctx.metadata.get("component"):
            message += f" [component={ctx.metadata['component']}]"

        return message

    def get_error_summary(self) -> dict[str, Any]:
        """Get error summary for monitoring."""
        return {
            "error_code": self.error_context.error_code.value,
            "error_category": self.error_context.error_category.value,
            "severity": self.error_context.severity.value,
            "message": self.error_context.message,
            "timestamp": self.error_context.timestamp,
            "operation": self.error_context.metadata.get("operation"),
            "component": self.error_context.metadata.get("component"),
            "has_original_exception": self.original_exception is not None,
        }

    def __str__(self) -> str:
        """String representation of the error."""
        return self._build_error_message()


# Error factory functions
def create_connection_limit_error(
    limit_type: str,
    limit_value: int | float,
    current_value: int | float,
    connection_id: str | None = None,
    peer_id: ID | None = None,
    message: str | None = None,
    metadata: dict[str, Any] | None = None,
) -> ResourceLimitExceededError:
    """Create a connection limit exceeded error."""
    error_code = ErrorCode.CONNECTION_LIMIT_EXCEEDED
    if limit_type == "pending":
        error_code = ErrorCode.CONNECTION_PENDING_LIMIT_EXCEEDED
    elif limit_type == "established":
        error_code = ErrorCode.CONNECTION_ESTABLISHED_LIMIT_EXCEEDED
    elif limit_type == "per_peer":
        error_code = ErrorCode.CONNECTION_PER_PEER_LIMIT_EXCEEDED
    elif limit_type == "total":
        error_code = ErrorCode.CONNECTION_TOTAL_LIMIT_EXCEEDED

    context = ErrorContext(
        error_code=error_code,
        error_category=ErrorCategory.CONNECTION_LIMIT,
        severity=ErrorSeverity.HIGH,
        message=message or f"Connection {limit_type} limit exceeded",
        resource_type="connection",
        connection_id=connection_id,
        peer_id=peer_id,
        limit_type=limit_type,
        limit_value=limit_value,
        current_value=current_value,
        limit_percentage=(
            (current_value / limit_value * 100) if limit_value > 0 else None
        ),
        metadata=metadata or {},
    )

    return ResourceLimitExceededError(context)


def create_memory_limit_error(
    limit_type: str,
    limit_value: int | float,
    current_value: int | float,
    system_memory_total: int | None = None,
    system_memory_available: int | None = None,
    process_memory_bytes: int | None = None,
    message: str | None = None,
    metadata: dict[str, Any] | None = None,
) -> SystemResourceError:
    """Create a memory limit exceeded error."""
    error_code = ErrorCode.MEMORY_LIMIT_EXCEEDED
    if limit_type == "process":
        error_code = ErrorCode.MEMORY_PROCESS_LIMIT_EXCEEDED
    elif limit_type == "system":
        error_code = ErrorCode.MEMORY_SYSTEM_LIMIT_EXCEEDED

    context = ErrorContext(
        error_code=error_code,
        error_category=ErrorCategory.MEMORY_LIMIT,
        severity=ErrorSeverity.CRITICAL,
        message=message or f"Memory {limit_type} limit exceeded",
        resource_type="memory",
        limit_type=limit_type,
        limit_value=limit_value,
        current_value=current_value,
        limit_percentage=(
            (current_value / limit_value * 100) if limit_value > 0 else None
        ),
        system_memory_total=system_memory_total,
        system_memory_available=system_memory_available,
        system_memory_percent=(
            (100 - (system_memory_available / system_memory_total * 100))
            if system_memory_total and system_memory_available
            else None
        ),
        process_memory_bytes=process_memory_bytes,
        process_memory_percent=(
            (process_memory_bytes / system_memory_total * 100)
            if process_memory_bytes and system_memory_total
            else None
        ),
        metadata=metadata or {},
    )

    return SystemResourceError(context)


def create_configuration_error(
    config_key: str,
    config_value: Any,
    message: str | None = None,
    metadata: dict[str, Any] | None = None,
) -> ConfigurationError:
    """Create a configuration error."""
    context = ErrorContext(
        error_code=ErrorCode.CONFIG_ERROR,
        error_category=ErrorCategory.CONFIG_ERROR,
        severity=ErrorSeverity.MEDIUM,
        message=message or f"Configuration error for {config_key}",
        metadata={
            **(metadata or {}),
            "config_key": config_key,
            "config_value": str(config_value),
        },
    )

    return ConfigurationError(context)


def create_operational_error(
    operation: str,
    component: str,
    message: str | None = None,
    metadata: dict[str, Any] | None = None,
) -> OperationalError:
    """Create an operational error."""
    context = ErrorContext(
        error_code=ErrorCode.OPERATIONAL_ERROR,
        error_category=ErrorCategory.OPERATIONAL_ERROR,
        severity=ErrorSeverity.MEDIUM,
        message=message or f"Operational error in {component} during {operation}",
        metadata={
            **(metadata or {}),
            "operation": operation,
            "component": component,
        },
    )

    return OperationalError(context)



"""
Rate limiter for resource management.

This module implements a comprehensive rate limiter that provides
multiple rate limiting strategies for production-ready resource
management.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
import time
from typing import Any

import multiaddr

from nanolib.peer.id import ID

from .token_bucket import TokenBucket, TokenBucketConfig


class RateLimitStrategy(Enum):
    """Rate limiting strategies."""

    TOKEN_BUCKET = "token_bucket"
    FIXED_WINDOW = "fixed_window"
    SLIDING_WINDOW = "sliding_window"
    ADAPTIVE = "adaptive"


class RateLimitScope(Enum):
    """Rate limiting scopes."""

    GLOBAL = "global"
    PER_PEER = "per_peer"
    PER_CONNECTION = "per_connection"
    PER_PROTOCOL = "per_protocol"
    PER_ENDPOINT = "per_endpoint"


@dataclass
class RateLimitConfig:
    """Configuration for rate limiter."""

    # Rate limiting parameters
    strategy: RateLimitStrategy = RateLimitStrategy.TOKEN_BUCKET
    scope: RateLimitScope = RateLimitScope.GLOBAL

    # Token bucket parameters (for TOKEN_BUCKET strategy)
    refill_rate: float = 10.0  # Tokens per second
    capacity: float = 100.0  # Maximum tokens in bucket
    initial_tokens: float = 0.0  # Initial tokens in bucket

    # Window parameters (for FIXED_WINDOW and SLIDING_WINDOW strategies)
    window_size_seconds: float = 60.0  # Window size in seconds
    max_requests_per_window: int = 100  # Maximum requests per window

    # Adaptive parameters (for ADAPTIVE strategy)
    base_rate: float = 10.0  # Base rate for adaptive limiting
    max_rate: float = 100.0  # Maximum rate for adaptive limiting
    min_rate: float = 1.0  # Minimum rate for adaptive limiting
    adaptation_factor: float = 0.1  # Adaptation factor

    # Burst handling
    allow_burst: bool = True  # Allow burst up to capacity
    burst_multiplier: float = 1.0  # Burst capacity multiplier

    # Time parameters
    time_window_seconds: float = 1.0  # Time window for rate calculation
    min_interval_seconds: float = 0.001  # Minimum interval between refills

    # Monitoring parameters
    enable_monitoring: bool = True  # Enable monitoring and statistics
    max_history_size: int = 1000  # Maximum history size for monitoring

    # Scope-specific parameters
    max_peers: int = 1000  # Maximum number of peers to track
    max_connections: int = 1000  # Maximum number of connections to track
    max_protocols: int = 100  # Maximum number of protocols to track

    def __post_init__(self) -> None:
        """Validate configuration parameters."""
        if self.refill_rate <= 0:
            raise ValueError("refill_rate must be positive")
        if self.capacity <= 0:
            raise ValueError("capacity must be positive")
        if self.initial_tokens < 0:
            raise ValueError("initial_tokens must be non-negative")
        if self.initial_tokens > self.capacity:
            raise ValueError("initial_tokens cannot exceed capacity")
        if self.window_size_seconds <= 0:
            raise ValueError("window_size_seconds must be positive")
        if self.max_requests_per_window <= 0:
            raise ValueError("max_requests_per_window must be positive")
        if self.base_rate <= 0:
            raise ValueError("base_rate must be positive")
        if self.max_rate <= self.base_rate:
            raise ValueError("max_rate must be greater than base_rate")
        if self.min_rate <= 0:
            raise ValueError("min_rate must be positive")
        if self.min_rate > self.base_rate:
            raise ValueError("min_rate must be less than or equal to base_rate")
        if self.adaptation_factor <= 0 or self.adaptation_factor > 1:
            raise ValueError("adaptation_factor must be between 0 and 1")
        if self.burst_multiplier < 1.0:
            raise ValueError("burst_multiplier must be >= 1.0")
        if self.time_window_seconds <= 0:
            raise ValueError("time_window_seconds must be positive")
        if self.min_interval_seconds <= 0:
            raise ValueError("min_interval_seconds must be positive")
        if self.max_history_size <= 0:
            raise ValueError("max_history_size must be positive")
        if self.max_peers <= 0:
            raise ValueError("max_peers must be positive")
        if self.max_connections <= 0:
            raise ValueError("max_connections must be positive")
        if self.max_protocols <= 0:
            raise ValueError("max_protocols must be positive")


@dataclass
class RateLimitStats:
    """Statistics for rate limiter."""

    # Current state
    current_rate: float
    allowed_requests: int
    denied_requests: int
    total_requests: int

    # Rate information
    configured_rate: float
    effective_rate: float
    burst_available: float

    # Scope information
    scope: RateLimitScope
    tracked_entities: int

    # Time information
    time_since_last_request: float
    next_available_time: float

    def to_dict(self) -> dict[str, Any]:
        """Convert statistics to dictionary."""
        return {
            "current_rate": self.current_rate,
            "allowed_requests": self.allowed_requests,
            "denied_requests": self.denied_requests,
            "total_requests": self.total_requests,
            "configured_rate": self.configured_rate,
            "effective_rate": self.effective_rate,
            "burst_available": self.burst_available,
            "scope": self.scope.value,
            "tracked_entities": self.tracked_entities,
            "time_since_last_request": self.time_since_last_request,
            "next_available_time": self.next_available_time,
        }


class RateLimiter:
    """
    Comprehensive rate limiter implementation.

    This class implements multiple rate limiting strategies including
    token bucket, fixed window, sliding window, and adaptive limiting.
    """

    def __init__(self, config: RateLimitConfig):
        """
        Initialize rate limiter.

        Args:
            config: Rate limiter configuration

        """
        self.config = config

        # Rate limiting state
        self._token_buckets: dict[str, TokenBucket] = {}
        self._window_counts: dict[str, list[float]] = {}
        self._last_request_times: dict[str, float] = {}

        # Statistics
        self._total_requests: int = 0
        self._allowed_requests: int = 0
        self._denied_requests: int = 0

        # Adaptive rate limiting
        self._current_rate: float = config.base_rate
        self._last_adaptation_time: float = time.time()

    def _get_entity_key(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        protocol: str | None = None,
        endpoint: str | multiaddr.Multiaddr | None = None,
    ) -> str:
        """
        Get entity key for rate limiting.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            protocol: Protocol name (optional)
            endpoint: Endpoint address (optional)

        Returns:
            Entity key for rate limiting

        """
        if self.config.scope == RateLimitScope.GLOBAL:
            return "global"
        elif self.config.scope == RateLimitScope.PER_PEER:
            return f"peer:{peer_id}" if peer_id else "unknown"
        elif self.config.scope == RateLimitScope.PER_CONNECTION:
            return f"conn:{connection_id}" if connection_id else "unknown"
        elif self.config.scope == RateLimitScope.PER_PROTOCOL:
            return f"proto:{protocol}" if protocol else "unknown"
        elif self.config.scope == RateLimitScope.PER_ENDPOINT:
            endpoint_str = str(endpoint) if endpoint else "unknown"
            return f"endpoint:{endpoint_str}"
        else:
            return "unknown"

    def _get_or_create_token_bucket(self, entity_key: str) -> TokenBucket:
        """
        Get or create token bucket for entity.

        Args:
            entity_key: Entity key for rate limiting

        Returns:
            Token bucket for entity

        """
        if entity_key not in self._token_buckets:
            # Check if we've reached the limit for this scope
            if len(self._token_buckets) >= self._get_max_entities():
                # Remove oldest entity (simple LRU)
                oldest_key = min(
                    self._token_buckets.keys(),
                    key=lambda k: self._last_request_times.get(k, 0),
                )
                del self._token_buckets[oldest_key]
                if oldest_key in self._last_request_times:
                    del self._last_request_times[oldest_key]

            # Create new token bucket
            bucket_config = TokenBucketConfig(
                refill_rate=self.config.refill_rate,
                capacity=self.config.capacity,
                initial_tokens=self.config.initial_tokens,
                allow_burst=self.config.allow_burst,
                burst_multiplier=self.config.burst_multiplier,
                time_window_seconds=self.config.time_window_seconds,
                min_interval_seconds=self.config.min_interval_seconds,
                enable_monitoring=self.config.enable_monitoring,
                max_history_size=self.config.max_history_size,
            )
            self._token_buckets[entity_key] = TokenBucket(bucket_config)

        return self._token_buckets[entity_key]

    def _get_max_entities(self) -> int:
        """Get maximum number of entities to track."""
        if self.config.scope == RateLimitScope.PER_PEER:
            return self.config.max_peers
        elif self.config.scope == RateLimitScope.PER_CONNECTION:
            return self.config.max_connections
        elif self.config.scope == RateLimitScope.PER_PROTOCOL:
            return self.config.max_protocols
        else:
            return 1  # Global scope

    def _update_adaptive_rate(self, allowed: bool, current_time: float) -> None:
        """
        Update adaptive rate based on request outcome.

        Args:
            allowed: Whether request was allowed
            current_time: Current timestamp

        """
        if self.config.strategy != RateLimitStrategy.ADAPTIVE:
            return

        # Calculate time since last adaptation
        time_since_adaptation = current_time - self._last_adaptation_time

        # Only adapt if enough time has passed
        if time_since_adaptation < self.config.min_interval_seconds:
            return

        # Adapt rate based on outcome
        if allowed:
            # Increase rate if request was allowed
            self._current_rate = min(
                self._current_rate * (1 + self.config.adaptation_factor),
                self.config.max_rate,
            )
        else:
            # Decrease rate if request was denied
            self._current_rate = max(
                self._current_rate * (1 - self.config.adaptation_factor),
                self.config.min_rate,
            )

        # Update last adaptation time
        self._last_adaptation_time = current_time

    def try_allow(
        self,
        tokens: float = 1.0,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        protocol: str | None = None,
        endpoint: str | multiaddr.Multiaddr | None = None,
        current_time: float | None = None,
    ) -> bool:
        """
        Try to allow a request based on rate limiting.

        Args:
            tokens: Number of tokens to consume
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            protocol: Protocol name (optional)
            endpoint: Endpoint address (optional)
            current_time: Current timestamp (optional)

        Returns:
            True if request is allowed, False otherwise

        """
        if current_time is None:
            current_time = time.time()

        # Get entity key
        entity_key = self._get_entity_key(peer_id, connection_id, protocol, endpoint)

        # Update last request time
        self._last_request_times[entity_key] = current_time

        # Update statistics
        self._total_requests += 1

        # Apply rate limiting strategy
        allowed = False

        if self.config.strategy == RateLimitStrategy.TOKEN_BUCKET:
            # Token bucket strategy
            bucket = self._get_or_create_token_bucket(entity_key)
            allowed = bucket.try_consume(tokens, current_time)

        elif self.config.strategy == RateLimitStrategy.FIXED_WINDOW:
            # Fixed window strategy
            window_start = current_time - self.config.window_size_seconds
            if entity_key not in self._window_counts:
                self._window_counts[entity_key] = []

            # Clean up old requests
            self._window_counts[entity_key] = [
                timestamp
                for timestamp in self._window_counts[entity_key]
                if timestamp > window_start
            ]

            # Check if we're within the limit
            if (
                len(self._window_counts[entity_key])
                < self.config.max_requests_per_window
            ):
                allowed = True
                self._window_counts[entity_key].append(current_time)

        elif self.config.strategy == RateLimitStrategy.SLIDING_WINDOW:
            # Sliding window strategy (similar to fixed window for simplicity)
            window_start = current_time - self.config.window_size_seconds
            if entity_key not in self._window_counts:
                self._window_counts[entity_key] = []

            # Clean up old requests
            self._window_counts[entity_key] = [
                timestamp
                for timestamp in self._window_counts[entity_key]
                if timestamp > window_start
            ]

            # Check if we're within the limit
            if (
                len(self._window_counts[entity_key])
                < self.config.max_requests_per_window
            ):
                allowed = True
                self._window_counts[entity_key].append(current_time)

        elif self.config.strategy == RateLimitStrategy.ADAPTIVE:
            # Adaptive strategy
            bucket = self._get_or_create_token_bucket(entity_key)
            # Update bucket rate based on current adaptive rate
            bucket.config.refill_rate = self._current_rate
            allowed = bucket.try_consume(tokens, current_time)

        # Update statistics
        if allowed:
            self._allowed_requests += 1
        else:
            self._denied_requests += 1

        # Update adaptive rate
        self._update_adaptive_rate(allowed, current_time)

        return allowed

    def allow(
        self,
        tokens: float = 1.0,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        protocol: str | None = None,
        endpoint: str | multiaddr.Multiaddr | None = None,
        current_time: float | None = None,
    ) -> None:
        """
        Allow a request based on rate limiting (raises exception if denied).

        Args:
            tokens: Number of tokens to consume
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            protocol: Protocol name (optional)
            endpoint: Endpoint address (optional)
            current_time: Current timestamp (optional)

        Raises:
            ValueError: If request is denied by rate limiting

        """
        if not self.try_allow(
            tokens, peer_id, connection_id, protocol, endpoint, current_time
        ):
            raise ValueError("Request denied by rate limiting")

    def get_stats(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        protocol: str | None = None,
        endpoint: str | multiaddr.Multiaddr | None = None,
        current_time: float | None = None,
    ) -> RateLimitStats:
        """
        Get current statistics for the rate limiter.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            protocol: Protocol name (optional)
            endpoint: Endpoint address (optional)
            current_time: Current timestamp (optional)

        Returns:
            RateLimitStats: Current statistics

        """
        if current_time is None:
            current_time = time.time()

        # Get entity key
        entity_key = self._get_entity_key(peer_id, connection_id, protocol, endpoint)

        # Get entity-specific statistics
        if entity_key in self._token_buckets:
            bucket_stats = self._token_buckets[entity_key].get_stats(current_time)
            current_rate = bucket_stats.effective_rate
            burst_available = bucket_stats.burst_available
        else:
            current_rate = 0.0
            burst_available = 0.0

        # Calculate time since last request
        time_since_last_request = 0.0
        if entity_key in self._last_request_times:
            time_since_last_request = (
                current_time - self._last_request_times[entity_key]
            )

        # Calculate next available time
        next_available_time = 0.0
        if self.config.strategy == RateLimitStrategy.TOKEN_BUCKET:
            if entity_key in self._token_buckets:
                bucket_stats = self._token_buckets[entity_key].get_stats(current_time)
                next_available_time = bucket_stats.next_refill_time

        return RateLimitStats(
            current_rate=current_rate,
            allowed_requests=self._allowed_requests,
            denied_requests=self._denied_requests,
            total_requests=self._total_requests,
            configured_rate=self.config.refill_rate,
            effective_rate=(
                self._allowed_requests / self._total_requests
                if self._total_requests > 0
                else 0.0
            ),
            burst_available=burst_available,
            scope=self.config.scope,
            tracked_entities=len(self._token_buckets),
            time_since_last_request=time_since_last_request,
            next_available_time=next_available_time,
        )

    def get_entity_stats(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        protocol: str | None = None,
        endpoint: str | multiaddr.Multiaddr | None = None,
    ) -> dict[str, Any]:
        """
        Get entity-specific statistics.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            protocol: Protocol name (optional)
            endpoint: Endpoint address (optional)

        Returns:
            Dictionary with entity-specific statistics

        """
        entity_key = self._get_entity_key(peer_id, connection_id, protocol, endpoint)

        if entity_key in self._token_buckets:
            bucket = self._token_buckets[entity_key]
            return bucket.get_stats().to_dict()
        else:
            return {"exists": False}

    def reset(self) -> None:
        """Reset the rate limiter to initial state."""
        self._token_buckets.clear()
        self._window_counts.clear()
        self._last_request_times.clear()
        self._total_requests = 0
        self._allowed_requests = 0
        self._denied_requests = 0
        self._current_rate = self.config.base_rate
        self._last_adaptation_time = time.time()

    def __str__(self) -> str:
        """String representation of rate limiter."""
        return (
            f"RateLimiter(strategy={self.config.strategy.value}, "
            f"scope={self.config.scope.value}, "
            f"rate={self.config.refill_rate:.2f}/s, "
            f"capacity={self.config.capacity:.2f})"
        )


# Factory functions
def create_rate_limiter(
    strategy: RateLimitStrategy = RateLimitStrategy.TOKEN_BUCKET,
    scope: RateLimitScope = RateLimitScope.GLOBAL,
    refill_rate: float = 10.0,
    capacity: float = 100.0,
    initial_tokens: float = 0.0,
    allow_burst: bool = True,
    burst_multiplier: float = 1.0,
    time_window_seconds: float = 1.0,
    min_interval_seconds: float = 0.001,
    enable_monitoring: bool = True,
    max_history_size: int = 1000,
    max_peers: int = 1000,
    max_connections: int = 1000,
    max_protocols: int = 100,
) -> RateLimiter:
    """
    Create a rate limiter.

    Args:
        strategy: Rate limiting strategy
        scope: Rate limiting scope
        refill_rate: Tokens per second
        capacity: Maximum tokens in bucket
        initial_tokens: Initial tokens in bucket
        allow_burst: Allow burst up to capacity
        burst_multiplier: Burst capacity multiplier
        time_window_seconds: Time window for rate calculation
        min_interval_seconds: Minimum interval between refills
        enable_monitoring: Enable monitoring and statistics
        max_history_size: Maximum history size for monitoring
        max_peers: Maximum number of peers to track
        max_connections: Maximum number of connections to track
        max_protocols: Maximum number of protocols to track

    Returns:
        RateLimiter: Configured rate limiter

    """
    config = RateLimitConfig(
        strategy=strategy,
        scope=scope,
        refill_rate=refill_rate,
        capacity=capacity,
        initial_tokens=initial_tokens,
        allow_burst=allow_burst,
        burst_multiplier=burst_multiplier,
        time_window_seconds=time_window_seconds,
        min_interval_seconds=min_interval_seconds,
        enable_monitoring=enable_monitoring,
        max_history_size=max_history_size,
        max_peers=max_peers,
        max_connections=max_connections,
        max_protocols=max_protocols,
    )
    return RateLimiter(config)


def create_global_rate_limiter(
    refill_rate: float = 10.0,
    capacity: float = 100.0,
    initial_tokens: float = 0.0,
) -> RateLimiter:
    """
    Create a global rate limiter.

    Args:
        refill_rate: Tokens per second
        capacity: Maximum tokens in bucket
        initial_tokens: Initial tokens in bucket

    Returns:
        RateLimiter: Configured global rate limiter

    """
    return create_rate_limiter(
        strategy=RateLimitStrategy.TOKEN_BUCKET,
        scope=RateLimitScope.GLOBAL,
        refill_rate=refill_rate,
        capacity=capacity,
        initial_tokens=initial_tokens,
    )


def create_per_peer_rate_limiter(
    refill_rate: float = 10.0,
    capacity: float = 100.0,
    initial_tokens: float = 0.0,
    max_peers: int = 1000,
) -> RateLimiter:
    """
    Create a per-peer rate limiter.

    Args:
        refill_rate: Tokens per second
        capacity: Maximum tokens in bucket
        initial_tokens: Initial tokens in bucket
        max_peers: Maximum number of peers to track

    Returns:
        RateLimiter: Configured per-peer rate limiter

    """
    # If bursting is allowed and no explicit initial tokens were provided,
    # start buckets full to enable immediate burst as in common token-bucket usage.
    effective_initial = initial_tokens if initial_tokens > 0.0 else capacity
    return create_rate_limiter(
        strategy=RateLimitStrategy.TOKEN_BUCKET,
        scope=RateLimitScope.PER_PEER,
        refill_rate=refill_rate,
        capacity=capacity,
        initial_tokens=effective_initial,
        max_peers=max_peers,
    )


def create_adaptive_rate_limiter(
    base_rate: float = 10.0,
    max_rate: float = 100.0,
    min_rate: float = 1.0,
    adaptation_factor: float = 0.1,
    scope: RateLimitScope = RateLimitScope.GLOBAL,
) -> RateLimiter:
    """
    Create an adaptive rate limiter.

    Args:
        base_rate: Base rate for adaptive limiting
        max_rate: Maximum rate for adaptive limiting
        min_rate: Minimum rate for adaptive limiting
        adaptation_factor: Adaptation factor
        scope: Rate limiting scope

    Returns:
        RateLimiter: Configured adaptive rate limiter

    """
    config = RateLimitConfig(
        strategy=RateLimitStrategy.ADAPTIVE,
        scope=scope,
        refill_rate=base_rate,
        capacity=max_rate,
        initial_tokens=0.0,
        base_rate=base_rate,
        max_rate=max_rate,
        min_rate=min_rate,
        adaptation_factor=adaptation_factor,
    )
    return RateLimiter(config)



"""
Asynchronous persistent peerstore implementation for py-libp2p.

This module provides an asynchronous persistent peerstore that stores peer data
in a datastore backend, similar to the pstoreds implementation in go-libp2p.
All operations are purely asynchronous using trio.
"""

from collections import defaultdict
from collections.abc import AsyncIterable, Sequence
import logging
import time
from typing import Any

from multiaddr import Multiaddr
import trio
from trio import MemoryReceiveChannel, MemorySendChannel

from nanolib.abc_async import IAsyncPeerStore
from nanolib.crypto.keys import KeyPair, PrivateKey, PublicKey
from nanolib.peer.envelope import Envelope
from nanolib.peer.id import ID
from nanolib.peer.peerdata import PeerData, PeerDataError
from nanolib.peer.peerinfo import PeerInfo
from nanolib.peer.peerstore import PeerRecordState, PeerStoreError

from ..datastore.base import IDatastore
from ..serialization import (
    SerializationError,
    deserialize_addresses,
    deserialize_envelope,
    deserialize_latency,
    deserialize_metadata,
    deserialize_protocols,
    deserialize_record_state,
    serialize_addresses,
    serialize_envelope,
    serialize_latency,
    serialize_metadata,
    serialize_protocols,
    serialize_record_state,
)

logger = logging.getLogger(__name__)


class AsyncPersistentPeerStore(IAsyncPeerStore):
    """
    Asynchronous persistent peerstore implementation that stores peer data
    in a datastore backend.

    This implementation follows the IAsyncPeerStore interface with purely
    asynchronous operations. All data is persisted to the datastore backend
    using async/await, similar to the pstoreds implementation in go-libp2p.
    """

    def __init__(
        self,
        datastore: IDatastore,
        max_records: int = 10000,
        sync_interval: float = 1.0,
        auto_sync: bool = True,
    ) -> None:
        """
        Initialize asynchronous persistent peerstore.

        Args:
            datastore: The asynchronous datastore backend to use for persistence
            max_records: Maximum number of peer records to store
            sync_interval: Minimum interval between sync operations (seconds)
            auto_sync: Whether to automatically sync after writes

        :raises ValueError: If datastore is None or max_records is invalid

        """
        if datastore is None:
            raise ValueError("datastore cannot be None")
        if max_records <= 0:
            raise ValueError("max_records must be positive")

        self.datastore = datastore
        self.max_records = max_records
        self.sync_interval = sync_interval
        self.auto_sync = auto_sync
        self._last_sync = time.time()
        self._pending_sync = False

        # In-memory caches for frequently accessed data
        self.peer_data_map: dict[ID, PeerData] = defaultdict(PeerData)
        self.addr_update_channels: dict[ID, MemorySendChannel[Multiaddr]] = {}
        self.peer_record_map: dict[ID, PeerRecordState] = {}
        self.local_peer_record: Envelope | None = None

        # Thread safety for concurrent access
        self._lock = trio.Lock()

        # Key prefixes for different data types
        self.ADDR_PREFIX = b"addr:"
        self.KEY_PREFIX = b"key:"
        self.METADATA_PREFIX = b"metadata:"
        self.PROTOCOL_PREFIX = b"protocol:"
        self.PEER_RECORD_PREFIX = b"peer_record:"
        self.LOCAL_RECORD_KEY = b"local_record"

    def _get_addr_key(self, peer_id: ID) -> bytes:
        """Get the datastore key for peer addresses."""
        return self.ADDR_PREFIX + peer_id.to_bytes()

    def _get_key_key(self, peer_id: ID) -> bytes:
        """Get the datastore key for peer keys."""
        return self.KEY_PREFIX + peer_id.to_bytes()

    def _get_metadata_key(self, peer_id: ID) -> bytes:
        """Get the datastore key for peer metadata."""
        return self.METADATA_PREFIX + peer_id.to_bytes()

    def _get_protocol_key(self, peer_id: ID) -> bytes:
        """Get the datastore key for peer protocols."""
        return self.PROTOCOL_PREFIX + peer_id.to_bytes()

    def _get_peer_record_key(self, peer_id: ID) -> bytes:
        """Get the datastore key for peer records."""
        return self.PEER_RECORD_PREFIX + peer_id.to_bytes()

    def _get_additional_key(self, peer_id: ID) -> bytes:
        """Get the datastore key for additional peer data fields."""
        return b"additional:" + peer_id.to_bytes()

    def _get_latency_key(self, peer_id: ID) -> bytes:
        """Get datastore key for peer latency data."""
        return b"latency:" + peer_id.to_bytes()

    async def _load_peer_data(self, peer_id: ID) -> PeerData:
        """Load peer data from datastore, creating if not exists."""
        async with self._lock:
            if peer_id not in self.peer_data_map:
                peer_data = PeerData()

                try:
                    # Load addresses
                    addr_key = self._get_addr_key(peer_id)
                    addr_data = await self.datastore.get(addr_key)
                    if addr_data:
                        peer_data.addrs = deserialize_addresses(addr_data)

                    # Load keys
                    key_key = self._get_key_key(peer_id)
                    key_data = await self.datastore.get(key_key)
                    if key_data:
                        # For now, store keys as metadata until keypair serialization
                        # keys_metadata = deserialize_metadata(key_data)
                        # TODO: Implement proper keypair deserialization
                        # peer_data.pubkey = deserialize_public_key(
                        #     keys_metadata.get(b"pubkey", b"")
                        # )
                        # peer_data.privkey = deserialize_private_key(
                        #     keys_metadata.get(b"privkey", b"")
                        # )
                        pass

                    # Load metadata
                    metadata_key = self._get_metadata_key(peer_id)
                    metadata_data = await self.datastore.get(metadata_key)
                    if metadata_data:
                        metadata_bytes = deserialize_metadata(metadata_data)
                        # Convert bytes back to appropriate types
                        peer_data.metadata = {
                            k: v.decode("utf-8") if isinstance(v, bytes) else v
                            for k, v in metadata_bytes.items()
                        }

                    # Load protocols
                    protocol_key = self._get_protocol_key(peer_id)
                    protocol_data = await self.datastore.get(protocol_key)
                    if protocol_data:
                        peer_data.protocols = deserialize_protocols(protocol_data)

                    # Load latency data
                    latency_key = self._get_latency_key(peer_id)
                    latency_data = await self.datastore.get(latency_key)
                    if latency_data:
                        latency_ns = deserialize_latency(latency_data)
                        # Convert nanoseconds back to seconds for latmap
                        peer_data.latmap = latency_ns / 1_000_000_000

                except (SerializationError, KeyError, ValueError, TypeError) as e:
                    logger.error(f"Failed to load peer data for {peer_id}: {e}")
                    # Continue with empty peer data
                except Exception:
                    logger.exception(
                        f"Unexpected error loading peer data for {peer_id}"
                    )
                    # Continue with empty peer data

                self.peer_data_map[peer_id] = peer_data

            return self.peer_data_map[peer_id]

    async def _save_peer_data(self, peer_id: ID, peer_data: PeerData) -> None:
        """
        Save peer data to datastore.

        :param peer_id: The peer ID to save data for
        :param peer_data: The peer data to save
        :raises PeerStoreError: If saving to datastore fails
        :raises SerializationError: If serialization fails
        """
        try:
            # Save addresses
            if peer_data.addrs:
                addr_key = self._get_addr_key(peer_id)
                addr_data = serialize_addresses(peer_data.addrs)
                await self.datastore.put(addr_key, addr_data)

            # Save keys (temporarily as metadata until proper keypair serialization)
            if peer_data.pubkey or peer_data.privkey:
                key_key = self._get_key_key(peer_id)
                keys_metadata = {}
                if peer_data.pubkey:
                    keys_metadata["pubkey"] = peer_data.pubkey.serialize()
                if peer_data.privkey:
                    keys_metadata["privkey"] = peer_data.privkey.serialize()
                key_data = serialize_metadata(keys_metadata)
                await self.datastore.put(key_key, key_data)

            # Save metadata
            if peer_data.metadata:
                metadata_key = self._get_metadata_key(peer_id)
                metadata_data = serialize_metadata(peer_data.metadata)
                await self.datastore.put(metadata_key, metadata_data)

            # Save protocols
            if peer_data.protocols:
                protocol_key = self._get_protocol_key(peer_id)
                protocol_data = serialize_protocols(peer_data.protocols)
                await self.datastore.put(protocol_key, protocol_data)

            # Save latency data if available
            if hasattr(peer_data, "latmap") and peer_data.latmap > 0:
                latency_key = self._get_latency_key(peer_id)
                # Convert seconds to nanoseconds for storage
                latency_ns = int(peer_data.latmap * 1_000_000_000)
                latency_data = serialize_latency(latency_ns)
                await self.datastore.put(latency_key, latency_data)

            # Conditionally sync to ensure data is persisted
            await self._maybe_sync()

        except SerializationError:
            raise
        except Exception as e:
            raise PeerStoreError(f"Failed to save peer data for {peer_id}") from e

    async def _maybe_sync(self) -> None:
        """
        Conditionally sync the datastore based on configuration.

        :raises PeerStoreError: If sync operation fails
        """
        if not self.auto_sync:
            return

        current_time = time.time()
        if current_time - self._last_sync >= self.sync_interval:
            try:
                await self.datastore.sync(b"")
                self._last_sync = current_time
                self._pending_sync = False
            except Exception as e:
                raise PeerStoreError("Failed to sync datastore") from e
        else:
            self._pending_sync = True

    async def _load_peer_record(self, peer_id: ID) -> PeerRecordState | None:
        """
        Load peer record from datastore.

        :param peer_id: The peer ID to load record for
        :return: PeerRecordState if found, None otherwise
        :raises PeerStoreError: If loading fails unexpectedly
        """
        async with self._lock:
            if peer_id not in self.peer_record_map:
                try:
                    record_key = self._get_peer_record_key(peer_id)
                    record_data = await self.datastore.get(record_key)
                    if record_data:
                        record_state = deserialize_record_state(record_data)
                        self.peer_record_map[peer_id] = record_state
                        return record_state
                except (SerializationError, KeyError, ValueError) as e:
                    logger.error(f"Failed to load peer record for {peer_id}: {e}")
                except Exception:
                    logger.exception(
                        f"Unexpected error loading peer record for {peer_id}"
                    )
            return self.peer_record_map.get(peer_id)

    async def _save_peer_record(
        self, peer_id: ID, record_state: PeerRecordState
    ) -> None:
        """
        Save peer record to datastore.

        :param peer_id: The peer ID to save record for
        :param record_state: The record state to save
        :raises PeerStoreError: If saving to datastore fails
        :raises SerializationError: If serialization fails
        """
        try:
            record_key = self._get_peer_record_key(peer_id)
            record_data = serialize_record_state(record_state)
            await self.datastore.put(record_key, record_data)
            self.peer_record_map[peer_id] = record_state
            await self._maybe_sync()
        except SerializationError:
            raise
        except Exception as e:
            raise PeerStoreError(f"Failed to save peer record for {peer_id}") from e

    async def _load_local_record(self) -> None:
        """
        Load local peer record from datastore.

        :raises PeerStoreError: If loading fails unexpectedly
        """
        try:
            local_data = await self.datastore.get(self.LOCAL_RECORD_KEY)
            if local_data:
                self.local_peer_record = deserialize_envelope(local_data)
        except (SerializationError, KeyError, ValueError) as e:
            logger.error(f"Failed to load local peer record: {e}")
        except Exception:
            logger.exception("Unexpected error loading local peer record")

    async def _save_local_record(self, envelope: Envelope) -> None:
        """
        Save local peer record to datastore.

        :param envelope: The envelope to save
        :raises PeerStoreError: If saving to datastore fails
        :raises SerializationError: If serialization fails
        """
        try:
            envelope_data = serialize_envelope(envelope)
            await self.datastore.put(self.LOCAL_RECORD_KEY, envelope_data)
            self.local_peer_record = envelope
            await self._maybe_sync()
        except SerializationError:
            raise
        except Exception as e:
            raise PeerStoreError("Failed to save local peer record") from e

    async def _clear_peerdata_from_datastore(self, peer_id: ID) -> None:
        """Clear peer data from datastore."""
        try:
            keys_to_delete = [
                self._get_addr_key(peer_id),
                self._get_key_key(peer_id),
                self._get_metadata_key(peer_id),
                self._get_protocol_key(peer_id),
                self._get_peer_record_key(peer_id),
                self._get_additional_key(peer_id),
            ]
            for key in keys_to_delete:
                await self.datastore.delete(key)
            await self.datastore.sync(b"")
        except Exception as e:
            logger.error(f"Failed to clear peer data from datastore for {peer_id}: {e}")

    # --------CORE ASYNC PEERSTORE METHODS--------

    async def get_local_record_async(self) -> Envelope | None:
        """Get the local-signed-record wrapped in Envelope"""
        if self.local_peer_record is None:
            await self._load_local_record()
        return self.local_peer_record

    async def set_local_record_async(self, envelope: Envelope) -> None:
        """Set the local-signed-record wrapped in Envelope"""
        await self._save_local_record(envelope)

    async def peer_info_async(self, peer_id: ID) -> PeerInfo:
        """
        :param peer_id: peer ID to get info for
        :return: peer info object
        """
        peer_data = await self._load_peer_data(peer_id)
        if peer_data.is_expired():
            peer_data.clear_addrs()
            await self._save_peer_data(peer_id, peer_data)
        return PeerInfo(peer_id, peer_data.get_addrs())

    async def peer_ids_async(self) -> list[ID]:
        """
        :return: all of the peer IDs stored in peer store
        """
        # Get all peer IDs from datastore by querying all prefixes
        peer_ids = set()

        try:
            # Query all address keys to find peer IDs
            for key, _ in self.datastore.query(self.ADDR_PREFIX):
                if key.startswith(self.ADDR_PREFIX):
                    peer_id_bytes = key[len(self.ADDR_PREFIX) :]
                    try:
                        peer_id = ID(peer_id_bytes)
                        peer_ids.add(peer_id)
                    except Exception:
                        continue  # Skip invalid peer IDs
        except Exception as e:
            logger.error(f"Failed to query peer IDs: {e}")

        # Also include any peer IDs from memory cache
        peer_ids.update(self.peer_data_map.keys())

        return list(peer_ids)

    async def clear_peerdata_async(self, peer_id: ID) -> None:
        """Clears all data associated with the given peer_id."""
        async with self._lock:
            # Remove from memory
            if peer_id in self.peer_data_map:
                del self.peer_data_map[peer_id]

            # Clear peer records from memory
            if peer_id in self.peer_record_map:
                del self.peer_record_map[peer_id]

            # Clear from datastore
            await self._clear_peerdata_from_datastore(peer_id)

    async def valid_peer_ids_async(self) -> list[ID]:
        """
        :return: all of the valid peer IDs stored in peer store
        """
        valid_peer_ids: list[ID] = []
        all_peer_ids = await self.peer_ids_async()

        for peer_id in all_peer_ids:
            try:
                peer_data = await self._load_peer_data(peer_id)
                if not peer_data.is_expired():
                    valid_peer_ids.append(peer_id)
                else:
                    peer_data.clear_addrs()
                    await self._save_peer_data(peer_id, peer_data)
            except Exception as e:
                logger.error(f"Error checking validity of peer {peer_id}: {e}")

        return valid_peer_ids

    async def _enforce_record_limit(self) -> None:
        """Enforce maximum number of stored records."""
        if len(self.peer_record_map) > self.max_records:
            # Remove oldest records based on sequence number
            sorted_records = sorted(
                self.peer_record_map.items(), key=lambda x: x[1].seq
            )
            records_to_remove = len(self.peer_record_map) - self.max_records
            for peer_id, _ in sorted_records[:records_to_remove]:
                # Remove from memory and datastore
                del self.peer_record_map[peer_id]
                try:
                    record_key = self._get_peer_record_key(peer_id)
                    await self.datastore.delete(record_key)
                except Exception as e:
                    logger.error(f"Failed to delete peer record for {peer_id}: {e}")

    async def start_cleanup_task(self, cleanup_interval: int = 3600) -> None:
        """Start periodic cleanup of expired peer records and addresses."""
        while True:
            await trio.sleep(cleanup_interval)
            await self._cleanup_expired_records()

    async def _cleanup_expired_records(self) -> None:
        """Remove expired peer records and addresses"""
        all_peer_ids = await self.peer_ids_async()
        expired_peers = []

        for peer_id in all_peer_ids:
            try:
                peer_data = await self._load_peer_data(peer_id)
                if peer_data.is_expired():
                    expired_peers.append(peer_id)
            except Exception as e:
                logger.error(f"Error checking expiry for peer {peer_id}: {e}")

        for peer_id in expired_peers:
            await self.maybe_delete_peer_record_async(peer_id)
            await self.clear_peerdata_async(peer_id)

        await self._enforce_record_limit()

    # --------PROTO-BOOK--------

    async def get_protocols_async(self, peer_id: ID) -> list[str]:
        """
        :param peer_id: peer ID to get protocols for
        :return: protocols (as list of strings)
        :raise PeerStoreError: if peer ID not found
        """
        peer_data = await self._load_peer_data(peer_id)
        return peer_data.get_protocols()

    async def add_protocols_async(self, peer_id: ID, protocols: Sequence[str]) -> None:
        """
        :param peer_id: peer ID to add protocols for
        :param protocols: protocols to add
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.add_protocols(list(protocols))
        await self._save_peer_data(peer_id, peer_data)

    async def set_protocols_async(self, peer_id: ID, protocols: Sequence[str]) -> None:
        """
        :param peer_id: peer ID to set protocols for
        :param protocols: protocols to set
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.set_protocols(list(protocols))
        await self._save_peer_data(peer_id, peer_data)

    async def remove_protocols_async(
        self, peer_id: ID, protocols: Sequence[str]
    ) -> None:
        """
        :param peer_id: peer ID to get info for
        :param protocols: unsupported protocols to remove
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.remove_protocols(protocols)
        await self._save_peer_data(peer_id, peer_data)

    async def supports_protocols_async(
        self, peer_id: ID, protocols: Sequence[str]
    ) -> list[str]:
        """
        :return: all of the peer IDs stored in peer store
        """
        peer_data = await self._load_peer_data(peer_id)
        return peer_data.supports_protocols(protocols)

    async def first_supported_protocol_async(
        self, peer_id: ID, protocols: Sequence[str]
    ) -> str:
        peer_data = await self._load_peer_data(peer_id)
        return peer_data.first_supported_protocol(protocols)

    async def clear_protocol_data_async(self, peer_id: ID) -> None:
        """Clears protocol data"""
        peer_data = await self._load_peer_data(peer_id)
        peer_data.clear_protocol_data()
        await self._save_peer_data(peer_id, peer_data)

    # ------METADATA---------

    async def get_async(self, peer_id: ID, key: str) -> Any:
        """
        :param peer_id: peer ID to get peer data for
        :param key: the key to search value for
        :return: value corresponding to the key
        :raise PeerStoreError: if peer ID or value not found
        """
        peer_data = await self._load_peer_data(peer_id)
        try:
            return peer_data.get_metadata(key)
        except PeerDataError as error:
            raise PeerStoreError() from error

    async def put_async(self, peer_id: ID, key: str, val: Any) -> None:
        """
        :param peer_id: peer ID to put peer data for
        :param key:
        :param value:
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.put_metadata(key, val)
        await self._save_peer_data(peer_id, peer_data)

    async def clear_metadata_async(self, peer_id: ID) -> None:
        """Clears metadata"""
        peer_data = await self._load_peer_data(peer_id)
        peer_data.clear_metadata()
        await self._save_peer_data(peer_id, peer_data)

    # -----CERT-ADDR-BOOK-----

    async def maybe_delete_peer_record_async(self, peer_id: ID) -> None:
        """
        Delete the signed peer record for a peer if it has no known
        (non-expired) addresses.
        """
        peer_data = await self._load_peer_data(peer_id)
        if not peer_data.get_addrs() and peer_id in self.peer_record_map:
            # Remove from memory and datastore
            del self.peer_record_map[peer_id]
            try:
                record_key = self._get_peer_record_key(peer_id)
                await self.datastore.delete(record_key)
                await self.datastore.sync(b"")
            except Exception as e:
                logger.error(f"Failed to delete peer record for {peer_id}: {e}")

    async def consume_peer_record_async(self, envelope: Envelope, ttl: int) -> bool:
        """
        Accept and store a signed PeerRecord, unless it's older than
        the one already stored.
        """
        record = envelope.record()
        peer_id = record.peer_id

        # Check if we have an existing record
        existing = await self._load_peer_record(peer_id)
        if existing and existing.seq > record.seq:
            return False

        # Store the new record
        new_addrs = set(record.addrs)
        record_state = PeerRecordState(envelope, record.seq)
        await self._save_peer_record(peer_id, record_state)

        # Update peer data
        peer_data = await self._load_peer_data(peer_id)
        peer_data.clear_addrs()
        peer_data.add_addrs(list(new_addrs))
        peer_data.set_ttl(ttl)
        peer_data.update_last_identified()
        await self._save_peer_data(peer_id, peer_data)

        return True

    async def get_peer_record_async(self, peer_id: ID) -> Envelope | None:
        """
        Retrieve the most recent signed PeerRecord `Envelope` for a peer, if it exists
        and is still relevant.
        """
        peer_data = await self._load_peer_data(peer_id)
        if not peer_data.is_expired() and peer_data.get_addrs():
            record_state = await self._load_peer_record(peer_id)
            if record_state is not None:
                return record_state.envelope
        return None

    # -------ADDR-BOOK--------

    async def add_addr_async(self, peer_id: ID, addr: Multiaddr, ttl: int) -> None:
        """
        :param peer_id: peer ID to add address for
        :param addr:
        :param ttl: time-to-live for the this record
        """
        await self.add_addrs_async(peer_id, [addr], ttl)

    async def add_addrs_async(
        self, peer_id: ID, addrs: Sequence[Multiaddr], ttl: int
    ) -> None:
        """
        :param peer_id: peer ID to add address for
        :param addrs:
        :param ttl: time-to-live for the this record
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.add_addrs(list(addrs))
        peer_data.set_ttl(ttl)
        peer_data.update_last_identified()
        await self._save_peer_data(peer_id, peer_data)

        # Notify address stream listeners
        if peer_id in self.addr_update_channels:
            for addr in addrs:
                try:
                    self.addr_update_channels[peer_id].send_nowait(addr)
                except trio.WouldBlock:
                    # Channel is full, skip this address update
                    # This is not a critical error as the address is already stored
                    pass

        await self.maybe_delete_peer_record_async(peer_id)

    async def addrs_async(self, peer_id: ID) -> list[Multiaddr]:
        """
        :param peer_id: peer ID to get addrs for
        :return: list of addrs of a valid peer.
        :raise PeerStoreError: if peer ID not found
        """
        peer_data = await self._load_peer_data(peer_id)
        if not peer_data.is_expired():
            return peer_data.get_addrs()
        else:
            peer_data.clear_addrs()
            await self._save_peer_data(peer_id, peer_data)
            raise PeerStoreError("peer ID is expired")

    async def clear_addrs_async(self, peer_id: ID) -> None:
        """
        :param peer_id: peer ID to clear addrs for
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.clear_addrs()
        await self._save_peer_data(peer_id, peer_data)
        await self.maybe_delete_peer_record_async(peer_id)

    async def peers_with_addrs_async(self) -> list[ID]:
        """
        :return: all of the peer IDs which has addrs stored in peer store
        """
        output: list[ID] = []
        all_peer_ids = await self.peer_ids_async()

        for peer_id in all_peer_ids:
            try:
                peer_data = await self._load_peer_data(peer_id)
                if len(peer_data.get_addrs()) >= 1:
                    if not peer_data.is_expired():
                        output.append(peer_id)
                    else:
                        peer_data.clear_addrs()
                        await self._save_peer_data(peer_id, peer_data)
            except Exception as e:
                logger.error(f"Error checking addresses for peer {peer_id}: {e}")

        return output

    async def addr_stream_async(self, peer_id: ID) -> AsyncIterable[Multiaddr]:  # type: ignore[override]
        """
        Returns an async stream of newly added addresses for the given peer.
        """
        send: MemorySendChannel[Multiaddr]
        receive: MemoryReceiveChannel[Multiaddr]

        send, receive = trio.open_memory_channel(0)
        self.addr_update_channels[peer_id] = send

        async for addr in receive:
            yield addr

    # -------KEY-BOOK---------

    async def add_pubkey_async(self, peer_id: ID, pubkey: PublicKey) -> None:
        """
        :param peer_id: peer ID to add public key for
        :param pubkey:
        :raise PeerStoreError: if peer ID and pubkey does not match
        """
        if ID.from_pubkey(pubkey) != peer_id:
            raise PeerStoreError("peer ID and pubkey does not match")
        peer_data = await self._load_peer_data(peer_id)
        peer_data.add_pubkey(pubkey)
        await self._save_peer_data(peer_id, peer_data)

    async def pubkey_async(self, peer_id: ID) -> PublicKey:
        """
        :param peer_id: peer ID to get public key for
        :return: public key of the peer
        :raise PeerStoreError: if peer ID or peer pubkey not found
        """
        peer_data = await self._load_peer_data(peer_id)
        try:
            return peer_data.get_pubkey()
        except PeerDataError as e:
            raise PeerStoreError("peer pubkey not found") from e

    async def add_privkey_async(self, peer_id: ID, privkey: PrivateKey) -> None:
        """
        :param peer_id: peer ID to add private key for
        :param privkey:
        :raise PeerStoreError: if peer ID or peer privkey not found
        """
        if ID.from_pubkey(privkey.get_public_key()) != peer_id:
            raise PeerStoreError("peer ID and privkey does not match")
        peer_data = await self._load_peer_data(peer_id)
        peer_data.add_privkey(privkey)
        await self._save_peer_data(peer_id, peer_data)

    async def privkey_async(self, peer_id: ID) -> PrivateKey:
        """
        :param peer_id: peer ID to get private key for
        :return: private key of the peer
        :raise PeerStoreError: if peer ID or peer privkey not found
        """
        peer_data = await self._load_peer_data(peer_id)
        try:
            return peer_data.get_privkey()
        except PeerDataError as e:
            raise PeerStoreError("peer privkey not found") from e

    async def add_key_pair_async(self, peer_id: ID, key_pair: KeyPair) -> None:
        """
        :param peer_id: peer ID to add private key for
        :param key_pair:
        """
        await self.add_pubkey_async(peer_id, key_pair.public_key)
        await self.add_privkey_async(peer_id, key_pair.private_key)

    async def peer_with_keys_async(self) -> list[ID]:
        """Returns the peer_ids for which keys are stored"""
        peer_ids_with_keys: list[ID] = []
        all_peer_ids = await self.peer_ids_async()

        for peer_id in all_peer_ids:
            try:
                peer_data = await self._load_peer_data(peer_id)
                if peer_data.pubkey is not None:
                    peer_ids_with_keys.append(peer_id)
            except Exception as e:
                logger.error(f"Error checking keys for peer {peer_id}: {e}")

        return peer_ids_with_keys

    async def clear_keydata_async(self, peer_id: ID) -> None:
        """Clears the keys of the peer"""
        peer_data = await self._load_peer_data(peer_id)
        peer_data.clear_keydata()
        await self._save_peer_data(peer_id, peer_data)

    # --------METRICS--------

    async def record_latency_async(self, peer_id: ID, RTT: float) -> None:
        """
        Records a new latency measurement for the given peer
        using Exponentially Weighted Moving Average (EWMA)
        """
        peer_data = await self._load_peer_data(peer_id)
        peer_data.record_latency(RTT)
        await self._save_peer_data(peer_id, peer_data)

    async def latency_EWMA_async(self, peer_id: ID) -> float:
        """
        :param peer_id: peer ID to get private key for
        :return: The latency EWMA value for that peer
        """
        peer_data = await self._load_peer_data(peer_id)
        return peer_data.latency_EWMA()

    async def clear_metrics_async(self, peer_id: ID) -> None:
        """Clear the latency metrics"""
        peer_data = await self._load_peer_data(peer_id)
        peer_data.clear_metrics()
        await self._save_peer_data(peer_id, peer_data)

    async def close_async(self) -> None:
        """Close the persistent peerstore and underlying datastore."""
        async with self._lock:
            # Close the datastore
            if hasattr(self.datastore, "close"):
                await self.datastore.close()

            # Clear memory caches
            self.peer_data_map.clear()
            self.peer_record_map.clear()
            self.local_peer_record = None

    async def __aenter__(self) -> "AsyncPersistentPeerStore":
        """Async context manager entry."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        """Async context manager exit."""
        await self.close_async()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: libp2p/crypto/pb/crypto.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'libp2p/crypto/pb/crypto.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dlibp2p/crypto/pb/crypto.proto\x12\tcrypto.pb\"?\n\tPublicKey\x12$\n\x08key_type\x18\x01 \x02(\x0e\x32\x12.crypto.pb.KeyType\x12\x0c\n\x04\x64\x61ta\x18\x02 \x02(\x0c\"@\n\nPrivateKey\x12$\n\x08key_type\x18\x01 \x02(\x0e\x32\x12.crypto.pb.KeyType\x12\x0c\n\x04\x64\x61ta\x18\x02 \x02(\x0c*S\n\x07KeyType\x12\x07\n\x03RSA\x10\x00\x12\x0b\n\x07\x45\x64\x32\x35\x35\x31\x39\x10\x01\x12\r\n\tSecp256k1\x10\x02\x12\t\n\x05\x45\x43\x44SA\x10\x03\x12\x0c\n\x08\x45\x43\x43_P256\x10\x04\x12\n\n\x06X25519\x10\x05')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.crypto.pb.crypto_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_KEYTYPE']._serialized_start=175
  _globals['_KEYTYPE']._serialized_end=258
  _globals['_PUBLICKEY']._serialized_start=44
  _globals['_PUBLICKEY']._serialized_end=107
  _globals['_PRIVATEKEY']._serialized_start=109
  _globals['_PRIVATEKEY']._serialized_end=173
# @@protoc_insertion_point(module_scope)
